"""Tests for the transfer of external data."""
import logging
from pathlib import Path

from housekeeper.store.models import Version

from cg.meta.transfer.external_data import ExternalDataAPI
from cg.store import Store
from cg.store.models import Sample
from cg.utils.checksum.checksum import check_md5sum, extract_md5sum
from tests.cli.workflow.conftest import dna_case
from tests.mocks.hk_mock import MockHousekeeperAPI
from tests.store.conftest import sample_obj


def test_create_log_dir(caplog, external_data_api: ExternalDataAPI, ticket_id: str):
    """Test generating the directory for logging."""
    caplog.set_level(logging.INFO)

    # WHEN the log directory is created
    log_dir = external_data_api.create_log_dir(ticket=ticket_id, dry_run=True)

    # THEN the path is not created since it is a dry run
    assert "Would have created path" in caplog.text

    # THEN the created path should start with 2 dirs and then the ticket id
    assert str(log_dir).startswith(f"/another/path/{ticket_id}")


def test_get_source_path(
    cust_sample_id: str,
    customer_id: str,
    external_data_api: ExternalDataAPI,
    ticket_id: str,
):
    """Test generating the source path."""
    # GIVEN a ticket number a customer and a customer sample id

    # WHEN the function is called and assigned
    source_path = external_data_api.get_source_path(
        ticket=ticket_id,
        customer=customer_id,
        cust_sample_id=cust_sample_id,
    )

    # THEN the return should be
    assert source_path == Path("server.name.se:/path/cust000/on/caesar/123456/child/")


def test_get_destination_path(
    customer_id: str,
    external_data_api: ExternalDataAPI,
    sample_id: str,
):
    """Test generating the destination path."""
    # GIVEN a customer and an internal sample id
    # WHEN the function creates the destination path
    destination_path = external_data_api.get_destination_path(
        customer=customer_id, lims_sample_id=sample_id
    )

    # THEN the destination path should contain the customer_id, ticket_id and sample_id
    assert destination_path == Path("/path/on/hasta/cust000/ADM1/")


def test_transfer_sample_files_from_source(
    caplog,
    customer_id: str,
    cust_sample_id: str,
    external_data_api: ExternalDataAPI,
    external_data_directory: Path,
    helpers,
    mocker,
    ticket_id: str,
):
    caplog.set_level(logging.INFO)

    # GIVEN a Store with three samples, where only two samples are present in the source folder
    for sample in [f"{cust_sample_id}1", f"{cust_sample_id}2", f"{cust_sample_id}3"]:
        helpers.add_sample(
            store=external_data_api.status_db, name=sample, original_ticket=ticket_id
        )

    mocker.patch.object(Store, "get_customer_id_from_ticket")
    Store.get_customer_id_from_ticket.return_value = customer_id

    mocker.patch.object(ExternalDataAPI, "get_source_path")
    external_data_api.get_source_path.return_value = external_data_directory

    external_data_api.source_path = str(Path("").joinpath(*external_data_directory.parts[:-2]))
    external_data_api.destination_path = str(
        Path("").joinpath(*external_data_directory.parts[:-1], "%s")
    )

    # WHEN the transfer is initiated
    external_data_api.transfer_sample_files_from_source(ticket=ticket_id, dry_run=True)

    # THEN only the two samples present in the source directory are included in the rsync

    assert str(external_data_directory) in caplog.text


def test_get_all_fastq(external_data_api: ExternalDataAPI, external_data_directory: Path):
    """Test the finding of fastq.gz files in customer directories"""
    # GIVEN a folder containing two folders with both fastq and md5 files
    for folder in external_data_directory.iterdir():
        # WHEN the list of file paths is created
        files = external_data_api.get_all_fastq(
            sample_folder=external_data_directory.joinpath(folder)
        )
        # THEN only fast.gz files are returned
        assert all(tmp.suffixes == [".fastq", ".gz"] for tmp in files)


def test_get_failed_fastq_paths(external_data_api: ExternalDataAPI, fastq_file: Path):
    bad_md5sum_file_path: Path = fastq_file.parent.joinpath("fastq_run_R1_001.fastq.gz")
    # GIVEN a list of paths with one fastq_file with a correct md5sum and one with an incorrect md5sum
    # When the failed paths are extracted
    failed_paths: list[Path] = external_data_api.get_failed_fastq_paths(
        [fastq_file, bad_md5sum_file_path]
    )
    # THEN only the path to the failed file should be in the list
    assert failed_paths == [bad_md5sum_file_path]


def test_add_files_to_bundles(
    external_data_api: ExternalDataAPI, fastq_file: Path, hk_version: Version, sample_id: str
):
    """Tests adding files to Housekeeper."""
    # GIVEN a file to be added

    # WHEN the files are added.
    external_data_api.add_files_to_bundles(
        fastq_paths=[fastq_file],
        last_version=hk_version,
        lims_sample_id=sample_id,
    )

    # THEN the function should return True and the file should have benn added.
    assert str(fastq_file.absolute()) in [idx.path for idx in hk_version.files]


def test_add_transfer_to_housekeeper(
    case_id,
    external_data_api: ExternalDataAPI,
    fastq_file: Path,
    mocker,
    ticket_id: str,
):
    """Test adding samples from a case to Housekeeper"""
    # GIVEN a Store with a DNA case, which is available for analysis
    case = external_data_api.status_db.get_case_by_internal_id(internal_id=case_id)
    mocker.patch.object(Store, "get_cases_by_ticket_id")
    Store.get_cases_by_ticket_id.return_value = [case]
    samples = [fam_sample.sample for fam_sample in case.links]

    # GIVEN a list of paths and only two samples being available
    mocker.patch.object(ExternalDataAPI, "get_all_paths")
    ExternalDataAPI.get_all_paths.return_value = [fastq_file]

    mocker.patch.object(MockHousekeeperAPI, "last_version")
    MockHousekeeperAPI.last_version.return_value = None

    mocker.patch.object(MockHousekeeperAPI, "get_files")
    MockHousekeeperAPI.get_files.return_value = []

    mocker.patch.object(Path, "iterdir")
    Path.iterdir.return_value = []

    mocker.patch.object(ExternalDataAPI, "get_available_samples")
    ExternalDataAPI.get_available_samples.return_value = samples[:-1]

    mocker.patch.object(Store, "cases")
    Store.cases.return_value = [{"internal_id": "yellowhog"}]

    mocker.patch.object(Store, "set_case_action")
    Store.set_case_action.return_value = None

    # THEN none of the samples should exist in housekeeper
    assert all(
        external_data_api.housekeeper_api.bundle(sample.internal_id) is None for sample in samples
    )

    # WHEN the sample bundles are added to housekeeper
    external_data_api.add_transfer_to_housekeeper(ticket=ticket_id)

    # THEN two sample bundles exist in housekeeper and the file has been added to those bundles bundles
    added_samples = list(external_data_api.housekeeper_api.bundles())
    assert all(
        sample.internal_id in [added_sample.name for added_sample in added_samples]
        for sample in samples[:-1]
    )
    assert all(
        sample.versions[0].files[0].path == str(fastq_file.absolute()) for sample in added_samples
    )
    # Then the sample that is not available should not exists
    assert samples[-1].internal_id not in [added_sample.name for added_sample in added_samples]


def test_get_available_samples(
    external_data_api: ExternalDataAPI,
    sample: Sample,
    ticket_id: str,
    tmpdir_factory,
):
    # GIVEN one such sample exists
    tmp_dir_path: Path = Path(tmpdir_factory.mktemp(sample.internal_id, numbered=False))
    available_samples = external_data_api.get_available_samples(
        folder=tmp_dir_path.parent, ticket=ticket_id
    )
    # THEN the function should return a list containing the sample object
    assert available_samples == [sample]
    tmp_dir_path.rmdir()


def test_curate_sample_folder(
    case_id, customer_id, external_data_api: ExternalDataAPI, tmpdir_factory
):
    case = external_data_api.status_db.get_case_by_internal_id(internal_id=case_id)
    sample: Sample = case.links[0].sample
    tmp_folder = Path(tmpdir_factory.mktemp(sample.name, numbered=False))
    external_data_api.curate_sample_folder(
        cust_name=customer_id, sample_folder=tmp_folder, force=False
    )
    assert (tmp_folder.parent / sample.internal_id).exists()
    assert not tmp_folder.exists()


def test_get_available_samples_no_samples_avail(
    external_data_api: ExternalDataAPI,
    ticket_id: str,
    tmpdir_factory,
):
    # GIVEN that the empty directory created does not contain any correct folders
    tmp_dir_path: Path = Path(tmpdir_factory.mktemp("not_sample_id", numbered=False))
    available_samples = external_data_api.get_available_samples(
        folder=tmp_dir_path, ticket=ticket_id
    )
    # THEN the function should return an empty list
    assert available_samples == []


def test_checksum(fastq_file: Path):
    """Tests if the function correctly calculates md5sum and returns the correct result."""
    # GIVEN a fastq file with corresponding correct md5 file and a fastq file with a corresponding incorrect md5 file
    bad_md5sum_file_path: Path = fastq_file.parent.joinpath("fastq_run_R1_001.fastq.gz")

    # THEN a file with a correct md5 sum should return true
    assert check_md5sum(file_path=fastq_file, md5sum="a95cbb265540a2261fce941059784fd1")

    # THEN a file with an incorrect md5 sum should return false
    assert not check_md5sum(
        file_path=bad_md5sum_file_path, md5sum="c690b0124173772ec4cbbc43709d84ee"
    )


def test_extract_checksum(fastq_file: Path):
    """Tests if the function successfully extract the correct md5sum."""

    # Given a file containing a md5sum
    md5sum_file = Path(f"{fastq_file.as_posix()}.md5")

    # Then the function should extract it
    assert extract_md5sum(md5sum_file=md5sum_file) == "a95cbb265540a2261fce941059784fd1"



"""Module for Loqusdb API."""
import logging
from pathlib import Path
from subprocess import CalledProcessError

from cg.constants.constants import FileFormat
from cg.exc import CaseNotFoundError, LoqusdbDeleteCaseError
from cg.io.controller import ReadStream
from cg.utils import Process
from cg.utils.dict import get_list_from_dictionary

LOG = logging.getLogger(__name__)


class LoqusdbAPI:
    """API for Loqusdb."""

    def __init__(self, binary_path: str, config_path: str):
        self.binary_path = binary_path
        self.config_path = config_path
        self.process = Process(binary=self.binary_path, config=self.config_path)

    def load(
        self,
        case_id: str,
        snv_vcf_path: Path,
        sv_vcf_path: Path | None = None,
        profile_vcf_path: Path | None = None,
        family_ped_path: Path | None = None,
        window_size: int | None = None,
        gq_threshold: int | None = None,
        qual_gq: bool | None = False,
        hard_threshold: float | None = None,
        soft_threshold: float | None = None,
    ) -> dict[str, int]:
        """Add observations to Loqusdb from VCF files."""
        load_params = {
            "--case-id": case_id,
            "--variant-file": snv_vcf_path.as_posix(),
            "--sv-variants": sv_vcf_path.as_posix() if sv_vcf_path else None,
            "--check-profile": profile_vcf_path.as_posix() if profile_vcf_path else None,
            "--family-file": family_ped_path.as_posix() if family_ped_path else None,
            "--max-window": str(window_size) if window_size else None,
            "--gq-threshold": str(gq_threshold) if gq_threshold else None,
            "--hard-threshold": str(hard_threshold) if hard_threshold else None,
            "--soft-threshold": str(soft_threshold) if soft_threshold else None,
        }
        load_call_params: list[str] = ["load"] + get_list_from_dictionary(load_params)
        load_call_params.append("--qual-gq") if qual_gq else None
        self.process.run_command(parameters=load_call_params)
        return self.get_nr_of_variants_in_file()

    def get_case(self, case_id: str) -> dict | None:
        """Return a case found in Loqusdb."""
        cases_parameters = ["cases", "-c", case_id, "--to-json"]
        self.process.run_command(parameters=cases_parameters)
        if not self.process.stdout:  # Case not in loqusdb, stdout of loqusdb command will be empty.
            LOG.info(f"Case {case_id} not found in {repr(self)}")
            return None

        return ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.process.stdout
        )[0]

    def get_duplicate(self, profile_vcf_path: Path, profile_threshold: float) -> dict | None:
        """Find matching profiles in Loqusdb."""
        duplicates_params = {
            "--check-vcf": profile_vcf_path.as_posix(),
            "--profile-threshold": str(profile_threshold),
        }
        duplicate_call_params: list = ["profile"] + get_list_from_dictionary(duplicates_params)

        try:
            self.process.run_command(parameters=duplicate_call_params)
        except CalledProcessError as exception:
            LOG.error(f"Could not execute the profile command for: {profile_vcf_path.as_posix()}")
            raise exception
        if not self.process.stdout:
            LOG.info(f"No duplicates found for profile: {profile_vcf_path.as_posix()}")
            return None

        return ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.process.stdout
        )

    def delete_case(self, case_id: str) -> None:
        """Remove a case from Loqusdb."""
        delete_call_parameters = ["delete", "-c", case_id]
        self.process.run_command(parameters=delete_call_parameters)
        for line in self.process.stderr_lines():
            if f"INFO Removing case {case_id}" in line:
                LOG.info(f"Removing case {case_id} from {repr(self)}")
                return
            if f"WARNING Case {case_id} does not exist" in line:
                LOG.error(f"Case {case_id} not found in {repr(self)}")
                raise CaseNotFoundError

        LOG.error(f"Could not delete case {case_id} from {repr(self)}")
        raise LoqusdbDeleteCaseError

    def get_nr_of_variants_in_file(self) -> dict[str, int]:
        """Return the number of variants in the uploaded to Loqusdb file."""
        nr_of_variants: int = 0
        for line in self.process.stderr_lines():
            line_content: str = line.split("INFO")[-1].strip()
            if "inserted" in line_content:
                nr_of_variants = int(line_content.split(":")[-1].strip())
        return {"variants": nr_of_variants}

    def __repr__(self):
        return f"LoqusdbAPI(binary_path={Path(self.binary_path).stem}, config_path={Path(self.config_path).stem})"



import logging
from datetime import datetime
from pathlib import Path

from cg.apps.housekeeper import models as hk_models
from cg.utils.commands import Process

from .models import CGDeliverables

LOG = logging.getLogger(__name__)


class HermesApi:
    """Class to communicate with hermes"""

    def __init__(self, config: dict):
        self.process = Process(binary=config["hermes"]["binary_path"])

    def convert_deliverables(
        self, deliverables_file: Path, pipeline: str, analysis_type: str | None = None
    ) -> CGDeliverables:
        """Convert deliverables file in raw pipeline format to CG format with hermes"""
        LOG.info("Converting pipeline deliverables to CG deliverables")
        convert_command = [
            "convert",
            "deliverables",
            "--pipeline",
            pipeline,
            str(deliverables_file),
        ]
        if analysis_type:
            convert_command.extend(["--analysis-type", analysis_type])
        self.process.run_command(convert_command)

        return CGDeliverables.model_validate_json(self.process.stdout)

    def create_housekeeper_bundle(
        self,
        bundle_name: str,
        deliverables: Path,
        pipeline: str,
        analysis_type: str | None,
        created: datetime | None,
    ) -> hk_models.InputBundle:
        """Convert pipeline deliverables to housekeeper bundle ready to be inserted into hk"""
        cg_deliverables: CGDeliverables = self.convert_deliverables(
            deliverables_file=deliverables, pipeline=pipeline, analysis_type=analysis_type
        )
        return self.get_housekeeper_bundle(
            deliverables=cg_deliverables, created=created, bundle_name=bundle_name
        )

    @staticmethod
    def get_housekeeper_bundle(
        deliverables: CGDeliverables, bundle_name: str, created: datetime | None = None
    ) -> hk_models.InputBundle:
        """Convert a deliverables object to a housekeeper object"""
        bundle_info = {
            "name": bundle_name,
            "files": [file_info.model_dump() for file_info in deliverables.files],
        }
        if created:
            bundle_info["created"] = created

        return hk_models.InputBundle(**bundle_info)



from pathlib import Path
from unittest import mock

import pytest
from housekeeper.store.models import File
from requests import HTTPError, Response

from cg.constants.archiving import ArchiveLocations
from cg.constants.constants import APIMethods
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.io.controller import APIRequest
from cg.meta.archive.archive import ARCHIVE_HANDLERS, FileAndSample, SpringArchiveAPI
from cg.meta.archive.ddn.constants import (
    FAILED_JOB_STATUSES,
    ONGOING_JOB_STATUSES,
    JobStatus,
)
from cg.meta.archive.ddn.ddn_data_flow_client import DDNDataFlowClient
from cg.meta.archive.ddn.models import (
    AuthToken,
    GetJobStatusPayload,
    GetJobStatusResponse,
    MiriaObject,
)
from cg.meta.archive.models import ArchiveHandler, FileTransferData
from cg.models.cg_config import DataFlowConfig
from cg.store.models import Sample


def test_add_samples_to_files(spring_archive_api: SpringArchiveAPI):
    """Tests matching Files to Samples when both files have a matching sample."""
    # GIVEN a list of SPRING Files to archive
    files_to_archive: list[
        File
    ] = spring_archive_api.housekeeper_api.get_non_archived_spring_files()

    # WHEN adding the Sample objects
    file_and_samples: list[FileAndSample] = spring_archive_api.add_samples_to_files(
        files_to_archive
    )

    # THEN each file should have a matching sample
    assert len(files_to_archive) == len(file_and_samples) > 0
    for file_and_sample in file_and_samples:
        # THEN the bundle name of each file should match the sample internal id
        assert file_and_sample.file.version.bundle.name == file_and_sample.sample.internal_id


def test_add_samples_to_files_missing_sample(spring_archive_api: SpringArchiveAPI):
    """Tests matching Files to Samples when one of the files does not match a Sample."""
    # GIVEN a list of SPRING Files to archive
    files_to_archive: list[
        File
    ] = spring_archive_api.housekeeper_api.get_non_archived_spring_files()
    # GIVEN one of the files does not match the
    files_to_archive[0].version.bundle.name = "does-not-exist"
    # WHEN adding the Sample objects
    file_and_samples: list[FileAndSample] = spring_archive_api.add_samples_to_files(
        files_to_archive
    )

    # THEN only one of the files should have a matching sample
    assert len(files_to_archive) != len(file_and_samples) > 0
    for file_and_sample in file_and_samples:
        # THEN the bundle name of each file should match the sample internal id
        assert file_and_sample.file.version.bundle.name == file_and_sample.sample.internal_id


def test_get_sample_exists(sample_id: str, spring_archive_api: SpringArchiveAPI):
    """Tests fetching a sample when the sample exists."""
    # GIVEN a sample that exists in the database
    file: File = spring_archive_api.housekeeper_api.get_files(bundle=sample_id).first()

    # WHEN getting the sample
    sample: Sample = spring_archive_api.get_sample(file)

    # THEN the correct sample should be returned
    assert sample.internal_id == sample_id


def test_get_sample_not_exists(
    caplog,
    spring_archive_api: SpringArchiveAPI,
    sample_id,
):
    """Tests fetching a sample when the sample does not exist."""
    # GIVEN a sample that does not exist in the database
    file: File = spring_archive_api.housekeeper_api.get_files(bundle=sample_id).first()
    sample_id: str = "non-existent-sample"
    file.version.bundle.name = sample_id

    # WHEN getting the sample
    sample: Sample = spring_archive_api.get_sample(file)

    # THEN the no sample should be returned
    # THEN both sample_id and file path should be logged
    assert not sample
    assert sample_id in caplog.text
    assert file.path in caplog.text


def test_convert_into_transfer_data(
    sample_id: str, spring_archive_api: SpringArchiveAPI, ddn_dataflow_config: DataFlowConfig
):
    """Tests instantiating the correct dataclass for a sample."""
    # GIVEN file and Sample
    file_and_sample = FileAndSample(
        file=spring_archive_api.housekeeper_api.get_files(bundle=sample_id).first(),
        sample=spring_archive_api.status_db.get_sample_by_internal_id(sample_id),
    )
    with mock.patch.object(
        DDNDataFlowClient,
        "_set_auth_tokens",
        return_value=123,
    ):
        # WHEN calling the corresponding archive method
        data_flow_client: ArchiveHandler = ARCHIVE_HANDLERS[ArchiveLocations.KAROLINSKA_BUCKET](
            config=ddn_dataflow_config
        )
    # WHEN using it to instantiate the correct class
    transferdata: list[FileTransferData] = data_flow_client.convert_into_transfer_data(
        files_and_samples=[file_and_sample],
    )

    # THEN the returned object should be of the correct type
    assert isinstance(transferdata[0], MiriaObject)


def test_call_corresponding_archiving_method(spring_archive_api: SpringArchiveAPI, sample_id: str):
    """Tests so that the correct archiving function is used when providing a Karolinska customer."""
    # GIVEN a file to be transferred
    # GIVEN a spring_archive_api with a mocked archive function
    file_and_sample = FileAndSample(
        file=spring_archive_api.housekeeper_api.get_files(bundle=sample_id).first(),
        sample=spring_archive_api.status_db.get_sample_by_internal_id(sample_id),
    )

    with mock.patch.object(
        DDNDataFlowClient,
        "_set_auth_tokens",
        return_value=123,
    ), mock.patch.object(
        DDNDataFlowClient,
        "archive_files",
        return_value=123,
    ) as mock_request_submitter:
        # WHEN calling the corresponding archive method
        spring_archive_api.archive_files_to_location(
            files_and_samples=[file_and_sample], archive_location=ArchiveLocations.KAROLINSKA_BUCKET
        )

    # THEN the correct archive function should have been called once
    mock_request_submitter.assert_called_once_with(files_and_samples=[file_and_sample])


@pytest.mark.parametrize("limit", [None, -1, 0, 1])
def test_archive_all_non_archived_spring_files(
    spring_archive_api: SpringArchiveAPI,
    caplog,
    ok_miria_response,
    archive_request_json,
    header_with_test_auth_token,
    test_auth_token: AuthToken,
    sample_id: str,
    limit: int | None,
):
    """Test archiving all non-archived SPRING files for Miria customers."""
    # GIVEN a populated status_db database with two customers, one DDN and one non-DDN,
    # with the DDN customer having two samples, and the non-DDN having one sample.

    # WHEN archiving all available files
    with mock.patch.object(
        AuthToken,
        "model_validate",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=ok_miria_response,
    ) as mock_request_submitter:
        spring_archive_api.archive_spring_files_and_add_archives_to_housekeeper(
            spring_file_count_limit=limit
        )

    # THEN the DDN archiving function should have been called with the correct destination and source if limit > 0
    if limit not in [0, -1]:
        mock_request_submitter.assert_called_with(
            api_method=APIMethods.POST,
            url="some/api/files/archive",
            headers=header_with_test_auth_token,
            json=archive_request_json,
            verify=False,
        )

        # THEN all spring files for Karolinska should have an entry in the Archive table in Housekeeper while no other
        # files should have an entry
        files: list[File] = spring_archive_api.housekeeper_api.files()
        for file in files:
            if SequencingFileTag.SPRING in [tag.name for tag in file.tags]:
                sample: Sample = spring_archive_api.status_db.get_sample_by_internal_id(
                    file.version.bundle.name
                )
                if sample and sample.archive_location == ArchiveLocations.KAROLINSKA_BUCKET:
                    assert file.archive
    else:
        mock_request_submitter.assert_not_called()


@pytest.mark.parametrize(
    "job_status, should_date_be_set",
    [
        (JobStatus.COMPLETED, True),
        (ONGOING_JOB_STATUSES[0], False),
        (FAILED_JOB_STATUSES[0], False),
    ],
)
def test_get_archival_status(
    spring_archive_api: SpringArchiveAPI,
    caplog,
    ok_miria_job_status_response,
    archive_request_json,
    header_with_test_auth_token,
    test_auth_token: AuthToken,
    archival_job_id: int,
    job_status: JobStatus,
    should_date_be_set: bool,
):
    # GIVEN a file with an ongoing archival
    file: File = spring_archive_api.housekeeper_api.files().first()
    spring_archive_api.housekeeper_api.add_archives(files=[file], archive_task_id=archival_job_id)

    # WHEN querying the task id and getting a "COMPLETED" response
    with mock.patch.object(
        AuthToken,
        "model_validate",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=ok_miria_job_status_response,
    ), mock.patch.object(
        GetJobStatusPayload,
        "get_job_status",
        return_value=GetJobStatusResponse(id=archival_job_id, status=job_status),
    ):
        spring_archive_api.update_ongoing_task(
            task_id=archival_job_id,
            archive_location=ArchiveLocations.KAROLINSKA_BUCKET,
            is_archival=True,
        )

    # THEN The Archive entry should have been updated
    if job_status == FAILED_JOB_STATUSES[0]:
        assert not file.archive
    else:
        assert bool(file.archive.archived_at) == should_date_be_set


@pytest.mark.parametrize(
    "job_status, should_date_be_set",
    [
        (JobStatus.COMPLETED, True),
        (ONGOING_JOB_STATUSES[0], False),
        (FAILED_JOB_STATUSES[0], False),
    ],
)
def test_get_retrieval_status(
    spring_archive_api: SpringArchiveAPI,
    caplog,
    ok_miria_job_status_response,
    archive_request_json,
    header_with_test_auth_token,
    archival_job_id: int,
    retrieval_job_id: int,
    test_auth_token,
    job_status,
    should_date_be_set,
):
    """Tests that the three different categories of retrieval statuses we have identified,
    i.e. failed, ongoing and successful, are handled correctly."""

    # GIVEN a file with an ongoing archival
    file: File = spring_archive_api.housekeeper_api.files().first()
    spring_archive_api.housekeeper_api.add_archives(files=[file], archive_task_id=archival_job_id)
    spring_archive_api.housekeeper_api.set_archive_retrieval_task_id(
        file_id=file.id, retrieval_task_id=retrieval_job_id
    )

    # WHEN querying the task id
    with mock.patch.object(
        AuthToken,
        "model_validate",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=ok_miria_job_status_response,
    ), mock.patch.object(
        GetJobStatusPayload,
        "get_job_status",
        return_value=GetJobStatusResponse(id=retrieval_job_id, status=job_status),
    ):
        spring_archive_api.update_ongoing_task(
            task_id=retrieval_job_id,
            archive_location=ArchiveLocations.KAROLINSKA_BUCKET,
            is_archival=False,
        )

    # THEN The Archive entry should have been updated
    if job_status == FAILED_JOB_STATUSES[0]:
        assert not file.archive.retrieval_task_id
    else:
        assert bool(file.archive.retrieved_at) == should_date_be_set


def test_retrieve_case(
    spring_archive_api: SpringArchiveAPI,
    caplog,
    ok_miria_response,
    trimmed_local_path,
    local_storage_repository,
    retrieve_request_json,
    header_with_test_auth_token,
    test_auth_token,
    archival_job_id: int,
    sample_with_spring_file: str,
):
    """Test retrieving all archived SPRING files tied to a sample for a Miria customer."""
    # GIVEN a populated status_db database with two customers, one DDN and one non-DDN,
    # with the DDN customer having two samples, and the non-DDN having one sample.
    files: list[File] = spring_archive_api.housekeeper_api.get_files(
        bundle=sample_with_spring_file, tags=[SequencingFileTag.SPRING]
    ).all()
    for file in files:
        spring_archive_api.housekeeper_api.add_archives(
            files=[file], archive_task_id=archival_job_id
        )
        assert not file.archive.retrieval_task_id
        assert file.archive

    sample: Sample = spring_archive_api.status_db.get_sample_by_internal_id(sample_with_spring_file)

    # WHEN archiving all available files
    with mock.patch.object(
        AuthToken,
        "model_validate",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=ok_miria_response,
    ) as mock_request_submitter:
        spring_archive_api.retrieve_case(sample.links[0].case.internal_id)

    retrieve_request_json["pathInfo"][0]["source"] += "/" + Path(files[0].path).name

    # THEN the DDN archiving function should have been called with the correct destination and source.
    mock_request_submitter.assert_called_with(
        api_method=APIMethods.POST,
        url="some/api/files/retrieve",
        headers=header_with_test_auth_token,
        json=retrieve_request_json,
        verify=False,
    )

    # THEN the Archive entry should have a retrieval task id set
    for file in files:
        assert file.archive.retrieval_task_id


def test_delete_file_raises_http_error(
    spring_archive_api: SpringArchiveAPI,
    failed_delete_file_response: Response,
    test_auth_token: AuthToken,
    archival_job_id: int,
):
    """Tests that an HTTP error is raised when the Miria response is unsuccessful for a delete file request,
    and that the file is not removed from Housekeeper."""

    # GIVEN a spring file which is archived via Miria
    spring_file: File = spring_archive_api.housekeeper_api.files(
        tags={SequencingFileTag.SPRING, ArchiveLocations.KAROLINSKA_BUCKET}
    ).first()
    spring_file_path: str = spring_file.path
    if not spring_file.archive:
        spring_archive_api.housekeeper_api.add_archives(
            files=[spring_file], archive_task_id=archival_job_id
        )
    spring_archive_api.housekeeper_api.set_archive_archived_at(
        file_id=spring_file.id, archiving_task_id=archival_job_id
    )

    # GIVEN that the request returns a failed response
    with mock.patch.object(
        DDNDataFlowClient,
        "_get_auth_token",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=failed_delete_file_response,
    ), pytest.raises(
        HTTPError
    ):
        # WHEN trying to delete the file via Miria and in Housekeeper

        # THEN an HTTPError should be raised
        spring_archive_api.delete_file(file_path=spring_file.path)

    # THEN the file should still be in Housekeeper
    assert spring_archive_api.housekeeper_api.files(path=spring_file_path)


def test_delete_file_success(
    spring_archive_api: SpringArchiveAPI,
    ok_delete_file_response: Response,
    test_auth_token: AuthToken,
    archival_job_id: int,
):
    """Tests that given a successful response from Miria, the file is deleted and removed from Housekeeper."""

    # GIVEN a spring file which is archived via Miria
    spring_file: File = spring_archive_api.housekeeper_api.files(
        tags={SequencingFileTag.SPRING, ArchiveLocations.KAROLINSKA_BUCKET}
    ).first()
    spring_file_id: int = spring_file.id
    if not spring_file.archive:
        spring_archive_api.housekeeper_api.add_archives(
            files=[spring_file], archive_task_id=archival_job_id
        )
    spring_archive_api.housekeeper_api.set_archive_archived_at(
        file_id=spring_file.id, archiving_task_id=archival_job_id
    )

    # GIVEN that the delete request returns a successful response
    with mock.patch.object(
        DDNDataFlowClient,
        "_get_auth_token",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=ok_delete_file_response,
    ):
        # WHEN trying to delete the file via Miria and in Housekeeper

        # THEN no error is raised
        spring_archive_api.delete_file(file_path=spring_file.path)

    # THEN the file is removed from Housekeeper
    assert not spring_archive_api.housekeeper_api.get_file(spring_file_id)



import http
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any
from unittest import mock

import pytest
from click.testing import CliRunner
from housekeeper.store.models import Bundle, File
from requests import Response

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.archiving import ArchiveLocations
from cg.constants.constants import DataDelivery, FileFormat, Pipeline
from cg.constants.subject import Sex
from cg.io.controller import WriteStream
from cg.meta.archive.archive import SpringArchiveAPI
from cg.meta.archive.ddn.constants import ROOT_TO_TRIM
from cg.meta.archive.ddn.ddn_data_flow_client import DDNDataFlowClient
from cg.meta.archive.ddn.models import AuthToken, MiriaObject, TransferPayload
from cg.meta.archive.models import FileAndSample
from cg.models.cg_config import CGConfig, DataFlowConfig
from cg.store import Store
from cg.store.models import Case, Customer, Sample
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="ddn_dataflow_config")
def ddn_dataflow_config(
    local_storage_repository: str, remote_storage_repository: str
) -> DataFlowConfig:
    """Returns a mock DDN Dataflow config."""
    return DataFlowConfig(
        database_name="test_db",
        user="test_user",
        password="DummyPassword",
        url=Path("some", "api", "url.com").as_posix(),
        local_storage=local_storage_repository,
        archive_repository=remote_storage_repository,
    )


@pytest.fixture
def ok_miria_response(ok_response: Response):
    ok_response._content = b'{"jobId": "123"}'
    return ok_response


@pytest.fixture
def ok_miria_job_status_response(ok_response: Response):
    ok_response._content = b'{"id": "123", "status": "Completed"}'
    return ok_response


@pytest.fixture
def archive_request_json(
    remote_storage_repository: str, local_storage_repository: str, trimmed_local_path: str
) -> dict:
    return {
        "osType": "Unix/MacOS",
        "createFolder": True,
        "pathInfo": [
            {
                "destination": f"{remote_storage_repository}ADM1",
                "source": local_storage_repository + trimmed_local_path,
            }
        ],
        "metadataList": [],
        "settings": [],
    }


@pytest.fixture
def retrieve_request_json(
    remote_storage_repository: str, local_storage_repository: str, trimmed_local_path: str
) -> dict[str, Any]:
    """Returns the body for a retrieval http post towards the DDN Miria API."""
    return {
        "osType": "Unix/MacOS",
        "createFolder": False,
        "pathInfo": [
            {
                "destination": local_storage_repository
                + Path(trimmed_local_path).parent.as_posix(),
                "source": f"{remote_storage_repository}ADM1",
            }
        ],
        "metadataList": [],
        "settings": [],
    }


@pytest.fixture
def header_with_test_auth_token() -> dict:
    return {
        "Content-Type": "application/json",
        "accept": "application/json",
        "Authorization": "Bearer test_auth_token",
    }


@pytest.fixture
def miria_auth_token_response(ok_response: Response):
    ok_response._content = b'{"access": "test_auth_token", "expire":15, "test_refresh_token":""}'
    return ok_response


@pytest.fixture
def test_auth_token() -> AuthToken:
    return AuthToken(
        access="test_auth_token",
        expire=int((datetime.now() + timedelta(minutes=20)).timestamp()),
        refresh="test_refresh_token",
    )


@pytest.fixture
def archival_job_id() -> int:
    return 123


@pytest.fixture
def retrieval_job_id() -> int:
    return 124


@pytest.fixture
def ddn_dataflow_client(ddn_dataflow_config: DataFlowConfig) -> DDNDataFlowClient:
    """Returns a DDNApi without tokens being set."""
    mock_ddn_auth_success_response = Response()
    mock_ddn_auth_success_response.status_code = 200
    mock_ddn_auth_success_response._content = WriteStream.write_stream_from_content(
        file_format=FileFormat.JSON,
        content={
            "access": "test_auth_token",
            "refresh": "test_refresh_token",
            "expire": int((datetime.now() + timedelta(minutes=20)).timestamp()),
        },
    ).encode()
    with mock.patch(
        "cg.meta.archive.ddn.ddn_data_flow_client.APIRequest.api_request_from_content",
        return_value=mock_ddn_auth_success_response,
    ):
        return DDNDataFlowClient(ddn_dataflow_config)


@pytest.fixture
def miria_file_archive(local_directory: Path, remote_path: Path) -> MiriaObject:
    """Return a MiriaObject for archiving."""
    return MiriaObject(source=local_directory.as_posix(), destination=remote_path.as_posix())


@pytest.fixture
def file_and_sample(spring_archive_api: SpringArchiveAPI, sample_id: str):
    return FileAndSample(
        file=spring_archive_api.housekeeper_api.get_files(bundle=sample_id).first(),
        sample=spring_archive_api.status_db.get_sample_by_internal_id(sample_id),
    )


@pytest.fixture
def trimmed_local_path(spring_archive_api: SpringArchiveAPI, sample_id: str):
    file: File = spring_archive_api.housekeeper_api.get_files(bundle=sample_id).first()
    return file.path[5:]


@pytest.fixture
def miria_file_retrieve(local_directory: Path, remote_path: Path) -> MiriaObject:
    """Return a MiriaObject for retrieval."""
    return MiriaObject(source=remote_path.as_posix(), destination=local_directory.as_posix())


@pytest.fixture
def transfer_payload(miria_file_archive: MiriaObject) -> TransferPayload:
    """Return a TransferPayload object containing two identical MiriaObject object."""
    return TransferPayload(
        files_to_transfer=[miria_file_archive, miria_file_archive.model_copy(deep=True)]
    )


@pytest.fixture
def remote_path() -> Path:
    """Returns a mock path."""
    return Path("/some", "place")


@pytest.fixture
def local_directory() -> Path:
    """Returns a mock path with /home as its root."""
    return Path(ROOT_TO_TRIM, "other", "place")


@pytest.fixture
def trimmed_local_directory(local_directory: Path) -> Path:
    """Returns the trimmed local directory."""
    return Path(f"/{local_directory.relative_to(ROOT_TO_TRIM)}")


@pytest.fixture
def local_storage_repository() -> str:
    """Returns a local storage repository."""
    return "local@storage:"


@pytest.fixture
def remote_storage_repository() -> str:
    """Returns a remote storage repository."""
    return "archive@repository:"


@pytest.fixture
def full_remote_path(remote_storage_repository: str, remote_path: Path) -> str:
    """Returns the merged remote repository and path."""
    return remote_storage_repository + remote_path.as_posix()


@pytest.fixture
def archive_store(
    base_store: Store,
    helpers: StoreHelpers,
    sample_id,
    father_sample_id,
    mother_sample_id,
    sample_name,
) -> Store:
    """Returns a store with samples for both a DDN customer and a non-DDN customer."""
    customer_ddn: Customer = base_store.add_customer(
        internal_id="CustWithDDN",
        invoice_address="Baker Street 221B",
        invoice_reference="Sherlock Holmes",
        name="Sherlock Holmes",
        is_clinical=True,
        data_archive_location=ArchiveLocations.KAROLINSKA_BUCKET,
    )
    customer_without_ddn: Customer = base_store.add_customer(
        internal_id="CustWithoutDDN",
        invoice_address="Wallaby Way 42",
        invoice_reference="P Sherman",
        name="P Sherman",
        is_clinical=False,
        data_archive_location="PDC",
    )

    new_samples: list[Sample] = [
        base_store.add_sample(
            name=sample_name,
            sex=Sex.MALE,
            internal_id=sample_id,
        ),
        base_store.add_sample(
            name="sample_2_with_ddn_customer",
            sex=Sex.MALE,
            internal_id=mother_sample_id,
        ),
        base_store.add_sample(
            name="sample_without_ddn_customer",
            sex=Sex.MALE,
            internal_id=father_sample_id,
        ),
    ]
    new_samples[0].customer = customer_ddn
    new_samples[1].customer = customer_ddn
    new_samples[2].customer = customer_without_ddn

    external_app = base_store.get_application_by_tag("WGXCUSC000").versions[0]
    wgs_app = base_store.get_application_by_tag("WGSPCFC030").versions[0]
    for sample in new_samples:
        sample.application_version = external_app if "external" in sample.name else wgs_app
    base_store.session.add(customer_ddn)
    base_store.session.add(customer_without_ddn)
    base_store.session.add_all(new_samples)
    base_store.session.commit()
    case: Case = base_store.add_case(
        data_analysis=Pipeline.MIP_DNA,
        data_delivery=DataDelivery.NO_DELIVERY,
        name="dummy_name",
        ticket="123",
        customer_id=customer_ddn.id,
    )
    base_store.relate_sample(case=case, sample=new_samples[0], status="unknown")
    base_store.session.add(case)
    base_store.session.commit()
    return base_store


@pytest.fixture
def sample_with_spring_file() -> str:
    return "ADM1"


@pytest.fixture
def spring_archive_api(
    populated_housekeeper_api: HousekeeperAPI,
    archive_store: Store,
    ddn_dataflow_config: DataFlowConfig,
    father_sample_id: str,
    sample_id: str,
    helpers,
) -> SpringArchiveAPI:
    """Returns an ArchiveAPI with a populated housekeeper store and a DDNDataFlowClient.
    Also adds /home/ as a prefix for each SPRING file for the DDNDataFlowClient to accept them."""
    populated_housekeeper_api.add_tags_if_non_existent(
        tag_names=[ArchiveLocations.KAROLINSKA_BUCKET]
    )
    for spring_file in populated_housekeeper_api.files(tags=[SequencingFileTag.SPRING]):
        spring_file.path = f"/home/{spring_file.path}"
        if spring_file.version.bundle.name == sample_id:
            spring_file.tags.append(
                populated_housekeeper_api.get_tag(name=ArchiveLocations.KAROLINSKA_BUCKET)
            )

    populated_housekeeper_api.commit()
    return SpringArchiveAPI(
        housekeeper_api=populated_housekeeper_api,
        status_db=archive_store,
        data_flow_config=ddn_dataflow_config,
    )


@pytest.fixture
def cli_runner() -> CliRunner:
    """Create a CliRunner"""
    return CliRunner()


@pytest.fixture
def base_context(
    base_store: Store, housekeeper_api: HousekeeperAPI, cg_config_object: CGConfig
) -> CGConfig:
    """context to use in CLI."""
    cg_config_object.status_db_ = base_store
    cg_config_object.housekeeper_api_ = housekeeper_api
    return cg_config_object


@pytest.fixture
def archive_context(
    base_context: CGConfig,
    real_housekeeper_api: HousekeeperAPI,
    path_to_spring_file_to_archive: str,
    path_to_spring_file_with_ongoing_archival: str,
    archival_job_id: int,
    helpers: StoreHelpers,
    ddn_dataflow_config: DataFlowConfig,
) -> CGConfig:
    base_context.housekeeper_api_ = real_housekeeper_api
    base_context.data_flow = ddn_dataflow_config

    customer = helpers.ensure_customer(
        store=base_context.status_db, customer_id="miria_customer", customer_name="Miriam"
    )
    customer.data_archive_location = ArchiveLocations.KAROLINSKA_BUCKET

    base_context.status_db.add_sample(
        name="sample_with_spring_files",
        sex=Sex.MALE,
        internal_id="sample_with_spring_files",
        **{"customer": "MiriaCustomer"},
    )
    helpers.add_sample(
        store=base_context.status_db, customer_id="miria_customer", internal_id="miria_sample"
    )
    bundle: Bundle = real_housekeeper_api.create_new_bundle_and_version(name="miria_sample")
    real_housekeeper_api.add_file(
        path=path_to_spring_file_to_archive,
        version_obj=bundle.versions[0],
        tags=[SequencingFileTag.SPRING, ArchiveLocations.KAROLINSKA_BUCKET],
    )
    file: File = real_housekeeper_api.add_file(
        path=path_to_spring_file_with_ongoing_archival,
        version_obj=bundle.versions[0],
        tags=[SequencingFileTag.SPRING, ArchiveLocations.KAROLINSKA_BUCKET],
    )
    file.id = 1234
    real_housekeeper_api.add_archives(files=[file], archive_task_id=archival_job_id)

    return base_context


@pytest.fixture
def path_to_spring_file_to_archive() -> str:
    return "/home/path/to/spring/file.spring"


@pytest.fixture
def path_to_spring_file_with_ongoing_archival() -> str:
    return "/home/path/to/ongoing/spring/file.spring"


@pytest.fixture
def failed_response() -> Response:
    response = Response()
    response.status_code = http.HTTPStatus.FORBIDDEN
    return response


@pytest.fixture
def failed_delete_file_response(failed_response: Response) -> Response:
    failed_response._content = b'{"detail":"Given token not valid for any token type","code":"token_not_valid","messages":[{"tokenClass":"AccessToken","tokenType":"access","message":"Token is invalid or expired"}]}'
    return failed_response


@pytest.fixture
def ok_delete_file_response(ok_response: Response) -> Response:
    ok_response._content = b'{"message":"Object has been deleted"}'
    return ok_response



import pytest

from cg.apps.sequencing_metrics_parser.sequencing_metrics_calculator import (
    calculate_average_quality_score,
    calculate_q30_bases_percentage,
)


@pytest.mark.parametrize(
    "q30_yield, total_yield, expected_result", [(100, 200, 0.5 * 100), (0, 0, 0)]
)
def test_calculate_bases_with_q30_percentage(
    q30_yield: int, total_yield: int, expected_result: float
):
    """Test calculating q30 ratio."""
    # GIVEN a number of bases with q30 and a total number of bases

    # WHEN calculating the q30 ratio
    q30_percentage = calculate_q30_bases_percentage(q30_yield=q30_yield, total_yield=total_yield)

    # THEN the q30 ratio should be the number of bases with q30 divided by the total number of bases
    assert q30_percentage == expected_result


@pytest.mark.parametrize(
    "total_quality_score, total_yield, expected_result", [(100, 200, 0.5), (0, 0, 0)]
)
def test_calculate_lane_mean_quality_score(
    total_quality_score: int, total_yield: int, expected_result: float
):
    """Test calculating the average quality score."""
    # GIVEN a total quality score and a total yield

    # WHEN calculating the average quality score
    avg_quality_score: float = calculate_average_quality_score(
        total_quality_score=total_quality_score, total_yield=total_yield
    )

    # THEN the average quality score should be the total quality score divided by the total yield
    assert avg_quality_score == expected_result



import logging
from glob import glob
from pathlib import Path

import click

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.tb import TrailblazerAPI
from cg.cli.demultiplex.copy_novaseqx_demultiplex_data import (
    hardlink_flow_cell_analysis_data,
    is_ready_for_post_processing,
    mark_as_demultiplexed,
    mark_flow_cell_as_queued_for_post_processing,
)
from cg.constants.demultiplexing import OPTION_BCL_CONVERTER, DemultiplexingDirsAndFiles
from cg.exc import FlowCellError
from cg.meta.demultiplex.utils import (
    create_manifest_file,
    is_flow_cell_sync_confirmed,
    is_manifest_file_required,
    is_syncing_complete,
)
from cg.models.cg_config import CGConfig
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData

LOG = logging.getLogger(__name__)

DRY_RUN = click.option("--dry-run", is_flag=True)


@click.command(name="all")
@click.option("--flow-cells-directory", type=click.Path(exists=True, file_okay=False))
@DRY_RUN
@click.pass_obj
def demultiplex_all(context: CGConfig, flow_cells_directory: click.Path, dry_run: bool):
    """Demultiplex all flow cells that are ready under the flow cells directory."""
    LOG.info("Running cg demultiplex all ...")
    demultiplex_api: DemultiplexingAPI = context.demultiplex_api
    demultiplex_api.set_dry_run(dry_run=dry_run)
    if flow_cells_directory:
        flow_cells_directory: Path = Path(str(flow_cells_directory))
    else:
        flow_cells_directory: Path = Path(demultiplex_api.flow_cells_dir)

    LOG.info(f"Search for flow cells ready to demultiplex in {flow_cells_directory}")
    for sub_dir in flow_cells_directory.iterdir():
        if not sub_dir.is_dir():
            continue
        LOG.info(f"Found directory {sub_dir}")
        try:
            flow_cell = FlowCellDirectoryData(flow_cell_path=sub_dir)
            LOG.info(f"Using {flow_cell.bcl_converter} for demultiplexing.")
        except FlowCellError:
            continue

        if not demultiplex_api.is_demultiplexing_possible(flow_cell=flow_cell):
            LOG.warning(f"Can not start demultiplexing for flow cell {flow_cell.id}!")
            continue

        if not flow_cell.validate_sample_sheet():
            LOG.warning(
                f"Malformed sample sheet. Run cg demultiplex samplesheet validate {flow_cell.sample_sheet_path}",
            )
            continue

        if not dry_run:
            demultiplex_api.prepare_output_directory(flow_cell)
            slurm_job_id: int = demultiplex_api.start_demultiplexing(flow_cell=flow_cell)
            tb_api: TrailblazerAPI = context.trailblazer_api
            demultiplex_api.add_to_trailblazer(
                tb_api=tb_api, slurm_job_id=slurm_job_id, flow_cell=flow_cell
            )


@click.command(name="flow-cell")
@click.argument("flow-cell-name")
@DRY_RUN
@OPTION_BCL_CONVERTER
@click.pass_obj
def demultiplex_flow_cell(
    context: CGConfig,
    dry_run: bool,
    flow_cell_name: str,
    bcl_converter: str,
):
    """Demultiplex a flow cell.

    flow cell name is the flow cell run directory name, e.g. '201203_D00483_0200_AHVKJCDRXX'
    """

    LOG.info(f"Running cg demultiplex flow cell, using {bcl_converter}")
    demultiplex_api: DemultiplexingAPI = context.demultiplex_api
    flow_cell_directory: Path = Path(context.demultiplex_api.flow_cells_dir, flow_cell_name)
    demultiplex_api.set_dry_run(dry_run=dry_run)
    LOG.info(f"setting flow cell id to {flow_cell_name}")
    LOG.info(f"setting demultiplexed runs dir to {demultiplex_api.demultiplexed_runs_dir}")

    try:
        flow_cell = FlowCellDirectoryData(
            flow_cell_path=flow_cell_directory, bcl_converter=bcl_converter
        )
    except FlowCellError as error:
        raise click.Abort from error

    if not demultiplex_api.is_demultiplexing_possible(flow_cell=flow_cell):
        LOG.warning("Can not start demultiplexing!")
        return

    if not flow_cell.validate_sample_sheet():
        LOG.warning(
            f"Malformed sample sheet. Run cg demultiplex samplesheet validate {flow_cell.sample_sheet_path}"
        )
        raise click.Abort

    if not dry_run:
        demultiplex_api.prepare_output_directory(flow_cell)
        slurm_job_id: int = demultiplex_api.start_demultiplexing(flow_cell=flow_cell)
        tb_api: TrailblazerAPI = context.trailblazer_api
        demultiplex_api.add_to_trailblazer(
            tb_api=tb_api, slurm_job_id=slurm_job_id, flow_cell=flow_cell
        )


@click.command(name="copy-completed-flow-cell")
@click.pass_obj
def copy_novaseqx_flow_cells(context: CGConfig):
    """Copy Novaseqx flow cells ready for post processing to demultiplexed runs."""
    flow_cells_dir: Path = Path(context.flow_cells_dir)
    demultiplexed_runs_dir: Path = Path(context.demultiplexed_flow_cells_dir)

    for flow_cell_dir in flow_cells_dir.iterdir():
        if is_ready_for_post_processing(
            flow_cell_dir=flow_cell_dir, demultiplexed_runs_dir=demultiplexed_runs_dir
        ):
            LOG.info(f"Copying {flow_cell_dir.name} to {demultiplexed_runs_dir}")
            hardlink_flow_cell_analysis_data(
                flow_cell_dir=flow_cell_dir, demultiplexed_runs_dir=demultiplexed_runs_dir
            )
            demultiplexed_runs_flow_cell_dir: Path = Path(
                demultiplexed_runs_dir, flow_cell_dir.name
            )
            mark_as_demultiplexed(demultiplexed_runs_flow_cell_dir)
            mark_flow_cell_as_queued_for_post_processing(flow_cell_dir)
        else:
            LOG.info(f"Flow cell {flow_cell_dir.name} is not ready for post processing, skipping.")


@click.command(name="confirm-flow-cell-sync")
@click.option(
    "--source-directory",
    required=True,
    help="The path from where the syncing is done.",
)
@click.pass_obj
def confirm_flow_cell_sync(context: CGConfig, source_directory: str):
    """Checks if all relevant files for the demultiplexing have been synced.
    If so it creates a CopyComplete.txt file to show that that is the case."""
    target_flow_cells_directory = Path(context.flow_cells_dir)
    for source_flow_cell in Path(source_directory).iterdir():
        target_flow_cell = Path(target_flow_cells_directory, source_flow_cell.name)
        if is_flow_cell_sync_confirmed(target_flow_cell):
            LOG.debug(f"Flow cell {source_flow_cell} has already been confirmed, skipping.")
            continue
        if is_syncing_complete(
            source_directory=source_flow_cell,
            target_directory=Path(target_flow_cells_directory, source_flow_cell.name),
        ):
            Path(
                target_flow_cells_directory,
                source_flow_cell.name,
                DemultiplexingDirsAndFiles.COPY_COMPLETE,
            ).touch()


@click.command(name="create-manifest-files")
@click.option(
    "--source-directory",
    required=True,
    help="The path from where the syncing is done.",
)
def create_manifest_files(source_directory: str):
    """Creates a file manifest for each flow cell in the source directory."""
    for source_flow_cell in glob(f"{source_directory}/*"):
        if is_manifest_file_required(Path(source_flow_cell)):
            create_manifest_file(Path(source_flow_cell))



"""Tests for cleaning FASTQ files."""
import logging
from pathlib import Path
from typing import Generator

from _pytest.logging import LogCaptureFixture
from housekeeper.store.models import File, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.archiving import PDC_ARCHIVE_LOCATION
from cg.meta.compress import files
from cg.models import CompressionData
from tests.cli.compress.conftest import MockCompressAPI
from tests.meta.compress.conftest import MockCompressionData
from tests.store_helpers import StoreHelpers


def test_remove_fastqs(
    compress_api: MockCompressAPI,
    compression_object: MockCompressionData,
    caplog: LogCaptureFixture,
):
    """Test remove_fastq method."""
    caplog.set_level(logging.DEBUG)

    # GIVEN existing FASTQ and flag file
    fastq_first: Path = compression_object.fastq_first
    fastq_second: Path = compression_object.fastq_second
    fastq_first.touch()
    fastq_second.touch()
    compression_object.spring_metadata_path.touch()

    # WHEN calling remove_fastq
    compress_api.remove_fastq(fastq_first=fastq_first, fastq_second=fastq_second)

    # THEN assert that the FASTQ-files are deleted
    assert not fastq_first.exists()
    assert not fastq_second.exists()

    # THEN assert that the flag file is still there since this holds important information
    assert compression_object.spring_metadata_path.exists()
    expected_output: str = f"Will remove {fastq_first} and {fastq_second}"
    assert expected_output in caplog.text
    assert f"FASTQ file {fastq_first} removed" in caplog.text
    assert f"FASTQ file {fastq_second} removed" in caplog.text


def test_update_hk_fastq(
    root_path: Path,
    real_housekeeper_api: Generator[HousekeeperAPI, None, None],
    compress_hk_fastq_bundle: dict,
    compression_files: MockCompressionData,
    compress_api: MockCompressAPI,
    helpers: StoreHelpers,
):
    """Test to update the FASTQ and SPRING paths in Housekeeper after completed compression."""

    # GIVEN real Housekeeper API populated with a Housekeeper bundle
    sample_id: str = compress_hk_fastq_bundle["name"]
    hk_api: Generator[HousekeeperAPI, None, None] = real_housekeeper_api
    helpers.ensure_hk_bundle(hk_api, compress_hk_fastq_bundle)
    compress_api.hk_api = hk_api

    # GIVEN that there are FASTQ files in Housekeeper
    hk_fastq_files: list = list(hk_api.files(tags=[SequencingFileTag.FASTQ]))
    assert hk_fastq_files

    # GIVEN that the SPRING files exist in disk but has been not added to Housekeeper
    assert compression_files.spring_file.exists()
    assert compression_files.spring_metadata_file.exists()
    hk_spring_files: list = list(hk_api.files(tags=[SequencingFileTag.SPRING]))
    hk_spring_metadata_files: list = list(hk_api.files(tags=[SequencingFileTag.SPRING_METADATA]))
    for spring_file in [hk_spring_files, hk_spring_metadata_files]:
        assert not spring_file

    # GIVEN a Housekeeper version and a compression object
    hk_version: Version = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    fastq: dict[str, dict] = files.get_fastq_files(sample_id=sample_id, version_obj=hk_version)
    run: str = list(fastq.keys())[0]
    compression: CompressionData = fastq[run]["compression_data"]

    # WHEN updating Housekeeper with compressed FASTQ files
    compress_api.update_fastq_hk(
        sample_id=sample_id,
        compression_obj=compression,
        hk_fastq_first=fastq[run]["hk_first"],
        hk_fastq_second=fastq[run]["hk_second"],
        archive_location=PDC_ARCHIVE_LOCATION,
    )

    # THEN assert that the SPRING files have been added to Housekeeper
    hk_spring_files: list[File] = list(real_housekeeper_api.files(tags=[SequencingFileTag.SPRING]))
    hk_spring_metadata_files: list[File] = list(
        real_housekeeper_api.files(tags=[SequencingFileTag.SPRING_METADATA])
    )

    # THEN assert that the Spring files and Spring metadata files have had the fastq file tags transferred
    for spring_file in [hk_spring_files, hk_spring_metadata_files]:
        assert spring_file
        for tag_name in [tag.name for tag in fastq[run]["hk_first"].tags]:
            if tag_name != SequencingFileTag.FASTQ:
                assert tag_name in spring_file.tags

    # THEN assert that the spring files have been tagged with the archive location
    for spring_file in hk_spring_files:
        assert PDC_ARCHIVE_LOCATION in [tag.name for tag in spring_file.tags]

    # THEN assert that the SPRING files have been added to bundles directory
    for spring_file in [hk_spring_files[0].path, hk_spring_metadata_files[0].path]:
        assert Path(root_path, spring_file).exists()

    # THEN assert that the FASTQ files are removed from Housekeeper
    hk_fastq_files: list[File] = list(hk_api.files(tags=[SequencingFileTag.FASTQ]))
    assert not hk_fastq_files


def test_cli_clean_fastqs_removed(
    populated_compress_fastq_api: MockCompressAPI,
    compression_files: MockCompressionData,
    sample: str,
):
    """Test to clean FASTQs after a successful FASTQ compression."""
    spring_file: Path = compression_files.spring_file
    spring_metadata_file: Path = compression_files.spring_metadata_file
    fastq_first: Path = compression_files.fastq_first_file
    fastq_second: Path = compression_files.fastq_second_file

    # GIVEN that the SPRING compression files exist
    assert spring_file.exists()
    assert spring_metadata_file.exists()

    # GIVEN that the FASTQ files exists
    assert fastq_first.exists()
    assert fastq_second.exists()

    # WHEN running the clean command
    populated_compress_fastq_api.clean_fastq(
        sample_id=sample, archive_location=PDC_ARCHIVE_LOCATION
    )

    # THEN assert SPRING files exists
    assert spring_file.exists()
    assert spring_metadata_file.exists()

    # THEN assert that the FASTQ files are removed
    assert not fastq_first.exists()
    assert not fastq_second.exists()


def test_cli_clean_fastqs_no_spring_metadata(
    populated_compress_fastq_api: MockCompressAPI,
    compression_files: MockCompressionData,
    sample: str,
):
    """Test to clean FASTQs when SPRING compression is not finished."""
    spring_file: Path = compression_files.spring_file
    spring_metadata_file: Path = compression_files.spring_metadata_path
    fastq_first: Path = compression_files.fastq_first_file
    fastq_second: Path = compression_files.fastq_second_file

    # GIVEN that the SPRING compression file exist
    assert spring_file.exists()

    # GIVEN that the SPRING metadata file does not exist
    assert not spring_metadata_file.exists()

    # WHEN running the clean command
    populated_compress_fastq_api.clean_fastq(
        sample_id=sample, archive_location=PDC_ARCHIVE_LOCATION
    )

    # THEN assert SPRING file exists
    assert spring_file.exists()

    # THEN assert that the FASTQ files are NOT removed
    assert fastq_first.exists()
    assert fastq_second.exists()


def test_cli_clean_fastqs_pending_compression_metadata(
    populated_compress_fastq_api: MockCompressAPI,
    compression_files: MockCompressionData,
    sample: str,
):
    """Test to clean FASTQs when SPRING compression is pending."""
    spring_file: Path = compression_files.spring_file
    fastq_first: Path = compression_files.fastq_first_file
    fastq_second: Path = compression_files.fastq_second_file

    # GIVEN that the SPRING compression file exist
    assert spring_file.exists()
    crunchy_flag_file = compression_files.pending_file
    assert crunchy_flag_file.exists()

    # WHEN running the clean command
    populated_compress_fastq_api.clean_fastq(
        sample_id=sample, archive_location=PDC_ARCHIVE_LOCATION
    )

    # THEN assert SPRING file exists
    assert spring_file.exists()

    # THEN assert that the FASTQ files are NOT removed
    assert fastq_first.exists()
    assert fastq_second.exists()



from sqlalchemy.orm import Query

from cg.constants import FlowCellStatus
from cg.store import Store
from cg.store.filters.status_flow_cell_filters import (
    filter_flow_cell_by_name,
    filter_flow_cell_by_name_search,
    filter_flow_cells_by_case,
    filter_flow_cells_with_statuses,
)
from cg.store.models import Case, Flowcell, Sample
from tests.store_helpers import StoreHelpers


def test_get_flow_cells_by_case(
    base_store: Store,
    case: Case,
    bcl2fastq_flow_cell_id: str,
    helpers: StoreHelpers,
    sample: Sample,
):
    """Test that a flow cell is returned when there is a flow cell with matching flow cell case."""

    # GIVEN a flow cell that exist in status db
    helpers.add_flow_cell(store=base_store, flow_cell_name=bcl2fastq_flow_cell_id, samples=[sample])

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell: list[Flowcell] | None = filter_flow_cells_by_case(
        flow_cells=base_store._get_join_flow_cell_sample_links_query(), case=case
    )

    # THEN a Flowcell type should be returned
    assert isinstance(returned_flow_cell[0], Flowcell)

    # THEN returned flow cell should have the same name as the flow cell in the database
    assert returned_flow_cell[0].name == bcl2fastq_flow_cell_id


def test_get_flow_cells_by_case_when_no_flow_cell_for_case(
    base_store: Store,
    case: Case,
):
    """Test that a flow cell is not returned when there is a flow cell with no matching flow cell for case."""

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell: list[Flowcell] | None = filter_flow_cells_by_case(
        flow_cells=base_store._get_join_flow_cell_sample_links_query(), case=case
    )

    # THEN returned flow cell should be the original flow cell
    assert not list(returned_flow_cell)


def test_get_flow_cell_by_id(base_store: Store, helpers: StoreHelpers, bcl2fastq_flow_cell_id: str):
    """Test that a flow cell is returned when there is a flow cell with matching flow cell id."""

    # GIVEN a flow cell that exist in status db
    flow_cell: Flowcell = helpers.add_flow_cell(
        store=base_store, flow_cell_name=bcl2fastq_flow_cell_id
    )

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell: Query = filter_flow_cell_by_name(
        flow_cells=base_store._get_query(table=Flowcell), flow_cell_name=bcl2fastq_flow_cell_id
    )

    # THEN returned flow cell should be the original flow cell
    assert isinstance(returned_flow_cell, Query)

    assert flow_cell is returned_flow_cell.first()


def test_get_flow_cell_by_id_and_by_enquiry(
    base_store: Store, helpers: StoreHelpers, bcl2fastq_flow_cell_id: str
):
    """Test that a flow cell is returned when there is a flow cell with enquiry flow cell id."""

    # GIVEN a flow cell that exist in status db
    flow_cell: Flowcell = helpers.add_flow_cell(
        store=base_store, flow_cell_name=bcl2fastq_flow_cell_id
    )

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell: list[Flowcell] = filter_flow_cell_by_name_search(
        flow_cells=base_store._get_query(table=Flowcell), name_search=bcl2fastq_flow_cell_id[:4]
    )

    # THEN a list of flow cells should be returned
    assert isinstance(returned_flow_cell[0], Flowcell)

    assert flow_cell is returned_flow_cell[0]


def test_get_flow_cells_with_statuses(
    base_store: Store, helpers: StoreHelpers, bcl2fastq_flow_cell_id: str
):
    """Test that a flow cell is returned when there is a flow cell with matching flow cell id."""

    # GIVEN a flow cell that exist in status db
    helpers.add_flow_cell(store=base_store, flow_cell_name=bcl2fastq_flow_cell_id)

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell_query: Query = filter_flow_cells_with_statuses(
        flow_cells=base_store._get_query(table=Flowcell),
        flow_cell_statuses=[FlowCellStatus.ON_DISK, FlowCellStatus.PROCESSING],
    )

    # THEN a query should be returned
    assert isinstance(returned_flow_cell_query, Query)


def test_filter_flow_cells_by_name(
    base_store: Store, helpers: StoreHelpers, bcl2fastq_flow_cell_id: str
):
    """Test flow cell is returned by name."""

    # GIVEN a flow cell that exist in status db
    flow_cell: Flowcell = helpers.add_flow_cell(
        store=base_store, flow_cell_name=bcl2fastq_flow_cell_id
    )

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell_query: Query = filter_flow_cell_by_name(
        flow_cells=base_store._get_query(table=Flowcell), flow_cell_name=flow_cell.name
    )
    returned_flow_cell: Flowcell = returned_flow_cell_query.first()

    # THEN a query should be returned
    assert isinstance(returned_flow_cell_query, Query)

    # THEN the flow cell should have the origin flow cell name
    assert returned_flow_cell.name == flow_cell.name



def calculate_q30_bases_percentage(q30_yield: int, total_yield: int) -> float:
    """
    Calculate the percentage of bases that have a Phred quality score of 30 or more.
    Args:
        q30_yield (int): The sum of all Q30 yields.
        total_yield (int): The total yield.
    Returns:
        float: Percentage of bases with Q30.
    """
    return round(q30_yield / total_yield, 2) * 100 if total_yield else 0


def calculate_average_quality_score(total_quality_score: int, total_yield: int) -> float:
    """
    Calculate the average quality score of all bases.
    Args:
        total_quality_score (int): The sum of quality scores.
        total_yield (int): The total yield.
    Returns:
        float: Average quality score of all bases.
    """
    return round(total_quality_score / total_yield, 2) if total_yield else 0



from sqlalchemy.orm import Query

from cg.store.api.core import Store
from cg.store.filters.status_panel_filters import filter_panel_by_abbrev
from cg.store.models import Panel


def test_filter_panel_by_abbrev_returns_correct_panel(store_with_panels: Store):
    """Test finding a panel by abbreviation when the abbreviation exists."""

    # GIVEN a store with multiple panels
    panel_query = store_with_panels._get_query(table=Panel)
    assert panel_query.count() > 1

    # Select a panel from the store
    panel = panel_query.first()
    assert isinstance(panel, Panel)

    # WHEN finding the panel by abbreviation
    filtered_panel_query: Query = filter_panel_by_abbrev(
        panels=panel_query,
        abbreviation=panel.abbrev,
    )

    # THEN the result is a query
    assert isinstance(filtered_panel_query, Query)

    # THEN the filtered panel is a Panel
    filtered_panel: Panel = filtered_panel_query.first()
    assert isinstance(filtered_panel, Panel)

    # THEN the filtered panel has the same abbreviation as the original panel
    assert filtered_panel.abbrev == panel.abbrev


def test_filter_panel_by_abbrev_returns_none_when_abbrev_does_not_exist(store_with_panels: Store):
    """Test finding a panel by abbreviation when the abbreviation does not exist in the store."""

    # GIVEN a store with panels
    panel_query: Query = store_with_panels._get_query(table=Panel)
    assert panel_query.count() > 0

    # WHEN finding a panel by an abbreviation that does not exist
    filtered_panels_query: Query = filter_panel_by_abbrev(
        panels=panel_query, abbreviation="nonexistent_abbrev"
    )

    # THEN the result is a query
    assert isinstance(filtered_panels_query, Query)

    # THEN no panels should be returned
    assert not filtered_panels_query.all()



import logging
from typing import Iterable

from housekeeper.store.models import File, Version

from cg.constants import (
    REQUIRED_APPLICATION_FIELDS,
    REQUIRED_CASE_FIELDS,
    REQUIRED_CUSTOMER_FIELDS,
    REQUIRED_DATA_ANALYSIS_MIP_DNA_FIELDS,
    REQUIRED_REPORT_FIELDS,
    REQUIRED_SAMPLE_METADATA_MIP_DNA_FIELDS,
    REQUIRED_SAMPLE_METADATA_MIP_DNA_WGS_FIELDS,
    REQUIRED_SAMPLE_METHODS_FIELDS,
    REQUIRED_SAMPLE_MIP_DNA_FIELDS,
    REQUIRED_SAMPLE_TIMESTAMP_FIELDS,
    Pipeline,
)
from cg.constants.scout import MIP_CASE_TAGS
from cg.meta.report.field_validators import get_million_read_pairs
from cg.meta.report.report_api import ReportAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.mip.mip_analysis import MipAnalysis
from cg.models.mip.mip_metrics_deliverables import get_sample_id_metric
from cg.models.report.metadata import MipDNASampleMetadataModel
from cg.models.report.report import CaseModel
from cg.models.report.sample import SampleModel
from cg.store.models import Case, Sample

LOG = logging.getLogger(__name__)


class MipDNAReportAPI(ReportAPI):
    """API to create Rare disease DNA delivery reports."""

    def __init__(self, config: CGConfig, analysis_api: MipDNAAnalysisAPI):
        super().__init__(config=config, analysis_api=analysis_api)
        self.analysis_api: MipDNAAnalysisAPI = analysis_api

    def get_sample_metadata(
        self, case: Case, sample: Sample, analysis_metadata: MipAnalysis
    ) -> MipDNASampleMetadataModel:
        """Return MIP DNA sample metadata to include in the report."""
        parsed_metrics = get_sample_id_metric(
            sample_id=sample.internal_id, sample_id_metrics=analysis_metadata.sample_id_metrics
        )
        sample_coverage: dict = self.get_sample_coverage(sample=sample, case=case)
        return MipDNASampleMetadataModel(
            bait_set=self.lims_api.capture_kit(lims_id=sample.internal_id),
            gender=parsed_metrics.predicted_sex,
            million_read_pairs=get_million_read_pairs(reads=sample.reads),
            mapped_reads=parsed_metrics.mapped_reads,
            mean_target_coverage=sample_coverage.get("mean_coverage"),
            pct_10x=sample_coverage.get("mean_completeness"),
            duplicates=parsed_metrics.duplicate_reads,
        )

    def get_sample_coverage(self, sample: Sample, case: Case) -> dict:
        """Return coverage values for a specific sample."""
        genes = self.get_genes_from_scout(panels=case.panels)
        sample_coverage = self.chanjo_api.sample_coverage(
            sample_id=sample.internal_id, panel_genes=genes
        )
        if sample_coverage:
            return sample_coverage
        LOG.warning(f"Could not calculate sample coverage for: {sample.internal_id}")
        return dict()

    def get_genes_from_scout(self, panels: list) -> list:
        """Return panel gene IDs information from Scout."""
        panel_genes = list()
        for panel in panels:
            panel_genes.extend(self.scout_api.get_genes(panel))
        panel_gene_ids = [gene.get("hgnc_id") for gene in panel_genes]
        return panel_gene_ids

    def get_genome_build(self, analysis_metadata: MipAnalysis) -> str:
        """Return build version of the genome reference of a specific case."""
        return analysis_metadata.genome_build

    def get_report_accreditation(
        self, samples: list[SampleModel], analysis_metadata: MipAnalysis = None
    ) -> bool:
        """Checks if the report is accredited or not by evaluating each of the sample process accreditations."""
        for sample in samples:
            if not sample.application.accredited:
                return False
        return True

    def get_required_fields(self, case: CaseModel) -> dict:
        """Return dictionary with the delivery report required fields for MIP DNA."""
        return {
            "report": REQUIRED_REPORT_FIELDS,
            "customer": REQUIRED_CUSTOMER_FIELDS,
            "case": REQUIRED_CASE_FIELDS,
            "applications": self.get_application_required_fields(
                case=case, required_fields=REQUIRED_APPLICATION_FIELDS
            ),
            "data_analysis": REQUIRED_DATA_ANALYSIS_MIP_DNA_FIELDS,
            "samples": self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_MIP_DNA_FIELDS
            ),
            "methods": self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_METHODS_FIELDS
            ),
            "timestamps": self.get_timestamp_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_TIMESTAMP_FIELDS
            ),
            "metadata": self.get_sample_metadata_required_fields(case=case),
        }

    @staticmethod
    def get_sample_metadata_required_fields(case: CaseModel) -> dict:
        """Return sample metadata required fields associated to a specific sample ID."""
        required_sample_metadata_fields = dict()
        for sample in case.samples:
            required_fields = (
                REQUIRED_SAMPLE_METADATA_MIP_DNA_WGS_FIELDS
                if "wgs" in sample.application.prep_category.lower()
                else REQUIRED_SAMPLE_METADATA_MIP_DNA_FIELDS
            )
            required_sample_metadata_fields.update({sample.id: required_fields})
        return required_sample_metadata_fields

    def get_template_name(self) -> str:
        """Return template name to render the delivery report."""
        return Pipeline.MIP_DNA + "_report.html"

    def get_upload_case_tags(self) -> dict:
        """Return MIP DNA upload case tags."""
        return MIP_CASE_TAGS

    def get_scout_uploaded_file_from_hk(self, case_id: str, scout_tag: str) -> str | None:
        """Return file path of the uploaded to Scout file given its tag."""
        version: Version = self.housekeeper_api.last_version(bundle=case_id)
        tags: list = self.get_hk_scout_file_tags(scout_tag=scout_tag)
        uploaded_files: Iterable[File] = self.housekeeper_api.get_files(
            bundle=case_id, tags=tags, version=version.id
        )
        if not tags or not any(uploaded_files):
            LOG.info(
                f"No files were found for the following Scout Housekeeper tag: {scout_tag} (case: {case_id})"
            )
            return None
        return uploaded_files[0].full_path



"""Model for down sampling meta data."""
import logging
from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Priority
from cg.store import Store
from cg.store.models import ApplicationVersion, Case, Sample
from cg.utils.calculations import multiply_by_million

LOG = logging.getLogger(__name__)


class DownsampleData:
    def __init__(
        self,
        status_db: Store,
        hk_api: HousekeeperAPI,
        sample_id: str,
        number_of_reads: float,
        case_id: str,
        case_name: str,
        out_dir: Path,
    ):
        """Initialize the downsample data and perform integrity checks."""
        self.status_db: Store = status_db
        self.housekeeper_api: HousekeeperAPI = hk_api
        self.sample_id: str = sample_id
        self.number_of_reads: float = number_of_reads
        self.case_id: str = case_id
        self.out_dir: Path = out_dir
        self.case_name: str = case_name
        self.original_sample: Sample = self.get_sample_to_downsample()
        self.original_case: Case = self.get_case_to_downsample()
        self.validate_enough_reads_to_downsample()
        self.downsampled_sample: Sample = self._generate_statusdb_downsampled_sample_record()
        self.downsampled_case: Case = self._generate_statusdb_downsampled_case()
        LOG.info(f"Downsample Data checks completed for {self.sample_id}")

    @property
    def downsampled_sample_name(
        self,
    ) -> str:
        """Return a new sample name with the number of reads to which it is down sampled in millions
        appended. DS stands for downsampled. The number of reads is converted to a string and the
        decimal point is replaced with a C."""
        return f"{self.sample_id}DS{self.convert_number_of_reads_to_string}M"

    @property
    def downsampled_case_name(
        self,
    ) -> str:
        """Return a case name with _downsampled appended."""
        return f"{self.case_name}_downsampled"

    @property
    def convert_number_of_reads_to_string(self) -> str:
        """Convert the number of reads to a string."""
        return str(self.number_of_reads).replace(".", "C")

    def get_sample_to_downsample(self) -> Sample:
        """
        Check if a sample exists in StatusDB.
        Raises:
            ValueError
        """
        sample: Sample = self.status_db.get_sample_by_internal_id(self.sample_id)
        if not sample:
            raise ValueError(f"Sample {self.sample_id} not found in StatusDB.")
        return sample

    def get_case_to_downsample(self) -> Case:
        """
        Check if a case exists in StatusDB.
            Raises: ValueError
        """
        case: Case = self.status_db.get_case_by_internal_id(self.case_id)
        if not case:
            raise ValueError(f"Case {self.case_id} not found in StatusDB.")
        return case

    def _generate_statusdb_downsampled_sample_record(
        self,
    ) -> Sample:
        """
        Generate a downsampled sample record for StatusDB.
        The new sample contains the original sample internal id and meta data.usDB
        """
        application_version: ApplicationVersion = self.get_application_version()
        downsampled_sample: Sample = self.status_db.add_sample(
            name=self.downsampled_sample_name,
            internal_id=self.downsampled_sample_name,
            sex=self.original_sample.sex,
            order=self.original_sample.order,
            reads=multiply_by_million(self.number_of_reads),
            downsampled_to=multiply_by_million(self.number_of_reads),
            from_sample=self.original_sample.internal_id,
            tumour=self.original_sample.is_tumour,
            priority=Priority.standard,
            customer=self.original_sample.customer,
            application_version=application_version,
        )
        return downsampled_sample

    def _generate_statusdb_downsampled_case(
        self,
    ) -> Case:
        """Generate a case for the downsampled samples. The new case uses existing case data."""
        if self.status_db.case_with_name_exists(case_name=self.downsampled_case_name):
            return self.status_db.get_case_by_name(self.downsampled_case_name)
        downsampled_case: Case = self.status_db.add_case(
            data_analysis=self.original_case.data_analysis,
            data_delivery=self.original_case.data_delivery,
            name=self.downsampled_case_name,
            panels=self.original_case.panels,
            priority=self.original_case.priority,
            ticket=self.original_case.latest_ticket,
        )
        downsampled_case.customer = self.original_case.customer
        return downsampled_case

    def validate_enough_reads_to_downsample(self) -> None:
        """
        Check if the sample has enough reads to downsample.
        Raises:
            ValueError
        """
        if multiply_by_million(self.number_of_reads) > self.original_sample.reads:
            raise ValueError(
                f"Sample {self.original_sample.internal_id} does not have enough reads ({self.original_sample.reads}) to down sample to "
                f"{self.number_of_reads}M."
            )

    @property
    def fastq_file_input_directory(self) -> Path:
        """Get the latest version directory for a sample in housekeeper."""
        LOG.debug(
            f"Trying to get input fastq directory for {self.original_sample.internal_id} from Housekeeper."
        )
        return self.housekeeper_api.get_latest_bundle_version(
            bundle_name=self.original_sample.internal_id,
        ).full_path

    @property
    def fastq_file_output_directory(self):
        """Get the output directory for the downsampled sample."""
        return Path(self.out_dir, self.downsampled_sample.internal_id)

    def get_sample_status(
        self,
    ) -> str:
        """Return the status of a sample."""
        return self.original_sample.links[0].status if self.original_sample.links else "unknown"

    def get_application_version(self) -> ApplicationVersion:
        """Return the application version for a sample."""
        return self.original_sample.application_version

    def create_down_sampling_working_directory(self) -> Path:
        """
        Create a working directory for the downsample job.
        Raises:
            FileExistsError
        """
        working_directory: Path = self.fastq_file_output_directory
        if working_directory.exists():
            raise FileExistsError(f"Working directory {working_directory} already exists.")
        LOG.info(f"Creating working directory {working_directory}")
        working_directory.mkdir(parents=True, exist_ok=True)
        return working_directory



"""Module for archiving and retrieving folders via DDN Dataflow."""
import logging
from datetime import datetime
from pathlib import Path
from urllib.parse import urljoin

from housekeeper.store.models import File
from requests import Response

from cg.constants.constants import APIMethods
from cg.exc import ArchiveJobFailedError, DdnDataflowAuthenticationError
from cg.io.controller import APIRequest
from cg.meta.archive.ddn.constants import (
    DESTINATION_ATTRIBUTE,
    FAILED_JOB_STATUSES,
    SOURCE_ATTRIBUTE,
    DataflowEndpoints,
    JobStatus,
)
from cg.meta.archive.ddn.models import (
    AuthPayload,
    AuthToken,
    DeleteFilePayload,
    GetJobStatusPayload,
    GetJobStatusResponse,
    MiriaObject,
    RefreshPayload,
    TransferPayload,
)
from cg.meta.archive.models import ArchiveHandler, FileAndSample
from cg.models.cg_config import DataFlowConfig

LOG = logging.getLogger(__name__)


class DDNDataFlowClient(ArchiveHandler):
    """Class for archiving and retrieving folders via DDN Dataflow."""

    def __init__(self, config: DataFlowConfig):
        self.database_name: str = config.database_name
        self.user: str = config.user
        self.password: str = config.password
        self.url: str = config.url
        self.archive_repository: str = config.archive_repository
        self.local_storage: str = config.local_storage
        self.auth_token: str
        self.refresh_token: str
        self.token_expiration: datetime
        self.headers: dict[str, str] = {
            "Content-Type": "application/json",
            "accept": "application/json",
        }
        self._set_auth_tokens()

    def _set_auth_tokens(self) -> None:
        """Retrieves and sets auth and refresh token from the REST-API."""
        auth_token: AuthToken = self._get_auth_token()
        self.refresh_token: str = auth_token.refresh
        self.auth_token: str = auth_token.access
        self.token_expiration: datetime = datetime.fromtimestamp(auth_token.expire)

    def _get_auth_token(self) -> AuthToken:
        """Retrieves auth and refresh token from the REST-API."""
        response: Response = APIRequest.api_request_from_content(
            api_method=APIMethods.POST,
            url=urljoin(base=self.url, url=DataflowEndpoints.GET_AUTH_TOKEN),
            headers=self.headers,
            json=AuthPayload(
                dbName=self.database_name,
                name=self.user,
                password=self.password,
            ).model_dump(),
            verify=False,
        )
        if not response.ok:
            raise DdnDataflowAuthenticationError(message=response.text)
        return AuthToken.model_validate(response.json())

    def _refresh_auth_token(self) -> None:
        """Updates the auth token by providing the refresh token to the REST-API."""
        auth_token: AuthToken = self._get_refreshed_auth_token()
        self.auth_token: str = auth_token.access
        self.token_expiration: datetime = datetime.fromtimestamp(auth_token.expire)

    def _get_refreshed_auth_token(self) -> AuthToken:
        response: Response = APIRequest.api_request_from_content(
            api_method=APIMethods.POST,
            url=urljoin(base=self.url, url=DataflowEndpoints.REFRESH_AUTH_TOKEN),
            headers=self.headers,
            json=RefreshPayload(refresh=self.refresh_token).model_dump(),
            verify=False,
        )
        return AuthToken.model_validate(response.json())

    @property
    def auth_header(self) -> dict[str, str]:
        """Returns an authorization header based on the current auth token, or updates it if
        needed."""
        if datetime.now() > self.token_expiration:
            self._refresh_auth_token()
        return {"Authorization": f"Bearer {self.auth_token}"}

    def archive_files(self, files_and_samples: list[FileAndSample]) -> int:
        """Archives all files provided, to their corresponding destination, as given by sources
        and destination in TransferData. Returns the job ID of the archiving task."""
        miria_file_data: list[MiriaObject] = self.convert_into_transfer_data(
            files_and_samples, is_archiving=True
        )
        archival_request: TransferPayload = self.create_transfer_request(
            miria_file_data=miria_file_data, is_archiving_request=True
        )
        return archival_request.post_request(
            headers=dict(self.headers, **self.auth_header),
            url=urljoin(base=self.url, url=DataflowEndpoints.ARCHIVE_FILES),
        ).job_id

    def retrieve_files(self, files_and_samples: list[FileAndSample]) -> int:
        """Retrieves the provided files and stores them in the corresponding sample bundle in
        Housekeeper."""
        miria_file_data: list[MiriaObject] = []
        for file_and_sample in files_and_samples:
            LOG.info(
                f"Will retrieve file {file_and_sample.file.path} for sample {file_and_sample.sample.internal_id} via Miria."
            )
            miria_object: MiriaObject = MiriaObject.create_from_file_and_sample(
                file=file_and_sample.file, sample=file_and_sample.sample, is_archiving=False
            )
            miria_file_data.append(miria_object)
        retrieval_request: TransferPayload = self.create_transfer_request(
            miria_file_data=miria_file_data, is_archiving_request=False
        )
        return retrieval_request.post_request(
            headers=dict(self.headers, **self.auth_header),
            url=urljoin(base=self.url, url=DataflowEndpoints.RETRIEVE_FILES),
        ).job_id

    def create_transfer_request(
        self, miria_file_data: list[MiriaObject], is_archiving_request: bool
    ) -> TransferPayload:
        """Performs the necessary curation of paths for the request to be valid, depending on if
        it is an archiving or a retrieve request.
        """
        source_prefix: str
        destination_prefix: str
        attribute: str

        source_prefix, destination_prefix, attribute = (
            (self.local_storage, self.archive_repository, SOURCE_ATTRIBUTE)
            if is_archiving_request
            else (self.archive_repository, self.local_storage, DESTINATION_ATTRIBUTE)
        )

        transfer_request = TransferPayload(
            files_to_transfer=miria_file_data, createFolder=is_archiving_request
        )
        transfer_request.trim_paths(attribute_to_trim=attribute)
        transfer_request.add_repositories(
            source_prefix=source_prefix, destination_prefix=destination_prefix
        )

        return transfer_request

    def convert_into_transfer_data(
        self, files_and_samples: list[FileAndSample], is_archiving: bool = True
    ) -> list[MiriaObject]:
        """Converts the provided files and samples to the format used for the request."""
        return [
            MiriaObject.create_from_file_and_sample(
                file=file_and_sample.file, sample=file_and_sample.sample, is_archiving=is_archiving
            )
            for file_and_sample in files_and_samples
        ]

    def is_job_done(self, job_id: int) -> bool:
        """Returns True if the specified job is completed, and false if it is still ongoing.
        Raises:
            ArchiveJobFailedError if the specified job has a failed status."""
        get_job_status_payload = GetJobStatusPayload(id=job_id)
        get_job_status_response: GetJobStatusResponse = get_job_status_payload.get_job_status(
            url=urljoin(self.url, DataflowEndpoints.GET_JOB_STATUS + str(job_id)),
            headers=dict(self.headers, **self.auth_header),
        )
        job_status: JobStatus = get_job_status_response.status
        LOG.info(f"Miria returned status {job_status} for job {job_id}")
        if job_status == JobStatus.COMPLETED:
            return True
        if job_status in FAILED_JOB_STATUSES:
            raise ArchiveJobFailedError(f"Job with id {job_id} failed with status {job_status}")
        return False

    @staticmethod
    def get_file_name(file: File) -> str:
        return Path(file.path).name

    def delete_file(self, file_and_sample: FileAndSample) -> None:
        """Deletes the given file via Miria."""
        file_name: str = self.get_file_name(file_and_sample.file)
        sample_id: str = file_and_sample.sample.internal_id
        delete_file_payload = DeleteFilePayload(
            global_path=f"{self.archive_repository}{sample_id}/{file_name}"
        )
        delete_file_payload.delete_file(
            url=urljoin(self.url, DataflowEndpoints.DELETE_FILE),
            headers=dict(self.headers, **self.auth_header),
        )



"""Tests for public methods in auto.py"""


def test_auto():
    """Test that create_app works"""
    # GIVEN all required configurations present in os.environ

    # WHEN app is imported from auto
    from cg.server.auto import app

    # THEN app exists
    assert app



"""Models used by hermes <-> cg interactions"""
import logging
from pathlib import Path

from pydantic import BaseModel, field_validator

from cg.exc import CgDataError

LOG = logging.getLogger(__name__)


class CGTag(BaseModel):
    """Specify the file output information from hermes"""

    path: str
    tags: list[str]
    mandatory: bool | None = False


class CGDeliverables(BaseModel):
    """Class that specifies the output from hermes"""

    pipeline: str
    bundle_id: str
    files: list[CGTag]

    @field_validator("files")
    @classmethod
    def remove_missing_files(cls, files: list[CGTag]) -> list[CGTag]:
        """Validates that the files in a suggested CGDeliverables object are correct.
        I.e. if a file doesn't exist an error is raised if the file was mandatory,
        otherwise it is simply removed from the list of files."""
        filtered_files: list[CGTag] = files.copy()
        for file in files:
            if file.mandatory and not Path(file.path).exists():
                raise CgDataError(f"Mandatory file cannot be found at {file.path}")
            if not Path(file.path).exists():
                LOG.info(f"Optional file {file.path} not found, removing from bundle.")
                filtered_files.remove(file)
        return filtered_files



"""Class to hold file information about a compression entity"""
import logging
import os
from datetime import datetime
from pathlib import Path

from cg.constants import FASTQ_FIRST_READ_SUFFIX, FASTQ_SECOND_READ_SUFFIX, FileExtensions
from cg.constants.compression import PENDING_PATH_SUFFIX

LOG = logging.getLogger(__name__)


class CompressionData:
    """Holds information about compression data"""

    def __init__(self, stub: Path = None):
        """Initialise a compression data object

        The stub is first part of the file name
        """
        self.stub = stub
        self.stub_string = str(self.stub)

    @property
    def pending_path(self) -> Path:
        """Return the path to a compression pending file"""
        return self.stub.with_suffix(PENDING_PATH_SUFFIX)

    @property
    def spring_path(self) -> Path:
        """Return the path to a SPRING file"""
        return self.stub.with_suffix(FileExtensions.SPRING)

    @property
    def encrypted_spring_path(self) -> Path:
        """Return the path to a SPRING file"""
        return self.stub.with_suffix(FileExtensions.SPRING).with_suffix(FileExtensions.GPG)

    @property
    def spring_metadata_path(self) -> Path:
        """Return the path to a SPRING metadata file"""
        return self.stub.with_suffix(".json")

    @property
    def analysis_dir(self) -> Path:
        """Return the path to folder where analysis is"""
        return self.stub.resolve().parent

    @property
    def fastq_first(self) -> Path:
        """Return the path to the first read in pair"""
        return Path(self.stub_string + FASTQ_FIRST_READ_SUFFIX)

    @property
    def fastq_second(self) -> Path:
        """Return the path to the second read in pair"""
        return Path(self.stub_string + FASTQ_SECOND_READ_SUFFIX)

    @property
    def run_name(self) -> str:
        """Return the name of the sequencing run identifier"""
        return self.stub.name

    def pair_exists(self) -> bool:
        """Check that both files in FASTQ pair exists"""
        LOG.info("Check if FASTQ pair exists")
        if not self.file_exists_and_is_accessible(self.fastq_first):
            return False
        return bool(self.file_exists_and_is_accessible(self.fastq_second))

    @staticmethod
    def is_absolute(file_path: Path) -> bool:
        """Check if file path can be resolved"""
        if not file_path.is_absolute():
            LOG.info("Could not resolve full path from HK to %s", file_path)
            return False
        return True

    @staticmethod
    def file_exists_and_is_accessible(file_path: Path) -> bool:
        """Check if file exists and is accesible"""
        try:
            if not file_path.exists():
                LOG.info("%s does not exist", file_path)
                return False
        except PermissionError:
            LOG.warning("Not permitted to access %s. Skipping", file_path)
            return False
        return True

    @staticmethod
    def is_symlink(file_path: Path) -> bool:
        """Check if file path is symbolik link"""
        LOG.info("Check if %s is a symlink", file_path)
        return os.path.islink(file_path)

    @staticmethod
    def get_change_date(file_path: Path) -> datetime:
        """Return the time when this file was changed"""
        changed_date = datetime.fromtimestamp(file_path.stat().st_mtime)
        LOG.info("File %s was changed %s", file_path, changed_date)
        return changed_date

    def spring_exists(self) -> bool:
        """Check if the SPRING file exists"""
        LOG.info("Check if SPRING archive file exists")
        return self.file_exists_and_is_accessible(self.spring_path)

    def metadata_exists(self) -> bool:
        """Check if the SPRING metadata file exists"""
        LOG.info("Check if SPRING metadata file exists")
        return self.file_exists_and_is_accessible(self.spring_metadata_path)

    def pending_exists(self) -> bool:
        """Check if the SPRING pending flag file exists"""
        LOG.info("Check if pending compression file exists")
        return self.file_exists_and_is_accessible(self.pending_path)

    def __str__(self):
        return f"CompressionData(run:{self.run_name})"

    def __repr__(self):
        return f"CompressionData(stub:{self.stub})"



"""Tests for the downsample command."""
from pathlib import Path

from click.testing import CliRunner, Result

from cg.cli.downsample import store_downsampled_samples
from cg.models.cg_config import CGConfig


def test_store_downsampled_fastq_files(
    downsample_context: CGConfig,
    downsample_sample_internal_id_1: str,
    downsample_sample_internal_id_2: str,
    cli_runner: CliRunner,
    downsample_dir: Path,
    tmp_path_factory,
):
    """Test to store the downsampled fastq files in housekeeper."""
    # GIVEN directories with downsampled sample
    for sample_id in [
        f"{downsample_sample_internal_id_1}_5M",
        f"{downsample_sample_internal_id_2}_5M",
    ]:
        Path(downsample_dir, f"{sample_id}").mkdir()
        Path(downsample_dir, f"{sample_id}", f"{sample_id}.fastq.gz").touch()
    # WHEN calling the cli command to store the fastq files
    result: Result = cli_runner.invoke(
        store_downsampled_samples,
        args=[f"{downsample_sample_internal_id_1}_5M", f"{downsample_sample_internal_id_2}_5M"],
        obj=downsample_context,
    )

    # THEN the command should run successfully
    assert result.exit_code == 0

    # THEN both samples should be stored in housekeeper
    for sample_id in [
        f"{downsample_sample_internal_id_1}_5M",
        f"{downsample_sample_internal_id_2}_5M",
    ]:
        assert downsample_context.housekeeper_api.get_latest_bundle_version(sample_id)
        assert downsample_context.housekeeper_api.get_file_from_latest_version(
            bundle_name=sample_id, tags=["fastq"]
        )
